#!/usr/local/bin/perl -w

# nsp2regex.pl version 0.3
# Program to create regex's from output of count.pl or statistic.pl
#
# Copyright (C) 2001-2003
# Ted Pedersen, University of Minnesota, Duluth
# tpederse@d.umn.edu
# Satanjeev Banerjee, University of Minnesota, Duluth
# bane0025@d.umn.edu
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
#                                 ChangeLog
# 
# May 10, 2003 TDP renamed bsp2regex as nsp2regex 
#
##############################################################################

#-----------------------------------------------------------------------------
#                              Start of program
#-----------------------------------------------------------------------------

# we have to use commandline options, so use the necessary package!
use Getopt::Long;

# first check if no commandline options have been provided... in which case
# print out the usage notes!
if ( $#ARGV == -1 )
{
    &minimalUsageNotes();
    exit;
}

# now get the options!
GetOptions("version", "help", "token=s", "sort" );

# if help has been requested, print out help!
if ( defined $opt_help )
{
    $opt_help = 1;
    &showHelp();
    exit;
}

# if version has been requested, show version!
if ( defined $opt_version )
{
    $opt_version = 1;
    &showVersion();
    exit;
}

# get the separator token!
if ( defined $opt_token )
{
    if ( !( -e $opt_token))
    {
	print "Cant find token definition file $opt_token.\n";
	askHelp();
	exit;
    }
    
    open TOKEN, $opt_token || die "Couldnt open $opt_token\n";
    
    while(<TOKEN>)
    {
	chomp;
	s/^\s*//g;
	s/\s*$//g;
	if (length($_) <= 0) { next; }
	if (!(/^\//) || !(/\/$/))
	{
	    print STDERR "Ignoring regex with no delimiters: $_\n";
	    next;
	}
	s/^\///;
	s/\/$//;
	push @tokenRegex, $_;
    }
    close TOKEN;
}
else 
{
    push @tokenRegex, "\\w+";
    push @tokenRegex, "[\.,;:\?!]";
}

# create the complete token regex
$tokenizerRegex = "";

foreach $token (@tokenRegex)
{
    if ( length($tokenizerRegex) > 0 ) 
    {
	$tokenizerRegex .= "|";
    }
    $tokenizerRegex .= "(";
    $tokenizerRegex .= $token;
    $tokenizerRegex .= ")";
}

# if you dont have any tokens to work with, abort
if ( $#tokenRegex < 0 ) 
{
    print "No token definitions to work with.\n";
    askHelp();
    exit;
}

$tokenizerRegex = "((<[^>]*>)*(" . $tokenizerRegex . ")(<[^>]*>)*\\s(<[^>]*>\\s)*)";

# now the meaty part!

while ( <> ) 
{
    chomp;
    
    if ( /^@/ && !(/^@@/) )
    {
	
	if ( $_ =~ /^@[^\.]*.WindowSize/ ) 
	{ 
	    s/^[^=]*=//;
	    $declaredWindowSize = $_;
	}
	next;
    }
    
    if ( /^@@/ ) { s/^@@/@/; }
    if (!(/<>/)) { next; }

    @words = split(/<>/);
    if ( $_ !~ /<>$/ ) { pop @words; }

    # if no words, then quit 
    if (!(@words)) { next; }

    # create the name using the <> 
    $name = join ("<>", @words);

    if (defined $declaredWindowSize) 
    {
	$windowSize = $declaredWindowSize - $#words - 1;
    }
    else 
    {
	$windowSize = 0;
    }

    # stick the window size in if windowSize > 0
    if ( $windowSize ) { $name .= "<>" . $windowSize; }
    
    # escape all the words 
    for ($i = 0; $i <= $#words; $i++)
    {
	$words[$i] = escape($words[$i]);
    }

    # wrap all the words in tags
    for ($i = 0; $i <= $#words; $i++)
    {
	$words[$i] = "(<[^>]*>)*" . $words[$i] . "(<[^>]*>)*";  
    }

    # create the separator
    $separator = "\\s(<[^>]*>\\s)*";  
    if ( $windowSize >= 1 )
    { 
	$separator .= $tokenizerRegex . "\{0,$windowSize\}";
    }

    $regex = join($separator, @words);
    $regex = "/\\s" . $regex . "\\s/";

    print "$regex \@name = $name\n";
}

# function to escape a punctuation mark, if necessary
sub escape
{
    my $param = shift;
    $param =~ s/([\\\/\|\(\)\[\]\{\}\^\$\*\+\?\.])/\\$1/g;
    return $param;
}

# function to output a minimal usage note when the user has not provided any
# commandline options
sub minimalUsageNotes
{
    print "Usage: nsp2regex.pl [OPTIONS] SOURCE [[, SOURCE] ...]\n";
    askHelp();
}

# function to output help messages for this program 
sub showHelp 
{
    print "Usage: nsp2regex.pl [OPTIONS] SOURCE [[, SOURCE] ...]\n\n";

    print "Converts n-grams in SOURCE to regular expressions. SOURCE must be the output of\n";
    print "count.pl or statistic.pl (both part of the Ngram Statistics Package).\n";
    print "Regular expressions are output to stdout.\n\n";

    print "OPTIONS:\n\n";

    print "  --version          Prints the version number.\n\n";

    print "  --help             Prints this help message.\n\n";

    print "  --token FILE       Uses tokens contained in FILE to create the separator\n";
    print "                     between tokens, when window size of SOURCE n-gram is\n";
    print "                     greater than the 'n' of the n-gram. Window sizes for\n";
    print "                     n-grams in SOURCE can be defined using the --extended\n";
    print "                     option in count.pl.\n\n";
}

# function to output the version number
sub showVersion
{
    print "nsp2regex.pl  -  Version 0.3\n";
    print "A component of the SenseTools Version 0.3\n";
    print "Copyright (C) 2001-2002, Ted Pedersen & Satanjeev Banerjee\n";
    print "Date of Last Update: May 10, 2003\n";
}

# function to output "ask for help" message when the user's goofed up!
sub askHelp
{
    print STDERR "Type nsp2regex.pl --help for help.\n";
}


