#!/usr/local/bin/perl -w

# xml2arff.pl version 0.3

# Program to take a set of regex's that is then matched to an xml file
# in order to create an arff representation of that xml file.
#
# Copyright (C) 2001-2003
# Ted Pedersen, University of Minnesota, Duluth
# tpederse@d.umn.edu
# Satanjeev Banerjee, University of Minnesota, Duluth
# bane0025@d.umn.edu
#
###############################################################################
#
#                       -------         CHANGELOG       ---------
#
#version        date            programmer      List of changes     change-id
#
# 0.3           01/17/2003      Amruta        	Changed few regexs
#						which assumed 	      ADP.3.1
#						Senseval instance
#						-id/lexelt formats.
#						Made more general and 
#						flexible regexes to 
#						allow internal spacing
#						in instance ids 
#
#		01/19/2003	Amruta		Changed the last      ADP.3.2 
#						attribute name to 
#						'senseclass' 
###############################################################################

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#-----------------------------------------------------------------------------
#                              Start of program
#-----------------------------------------------------------------------------

# we have to use commandline options, so use the necessary package!
use Getopt::Long;

# first check if no commandline options have been provided... in which case
# print out the usage notes!
if ( $#ARGV == -1 )
{
    &minimalUsageNotes();
    exit;
}

# now get the options!
GetOptions("version", "help", "relation=s", "training=s", "test=s");

# if help has been requested, print out help!
if ( defined $opt_help )
{
    $opt_help = $opt_help;
    &showHelp();
    exit;
}

# if version has been requested, show version!
if ( $opt_version )
{
    $opt_version = $opt_version;
    &showVersion();
    exit;
}

# get the relation name if provided, otherwise use default RELATION
$relationName = (defined $opt_relation) ? $opt_relation : "RELATION";

# check if both -training and -test provided or none!
if ( defined $opt_training xor defined $opt_test ) 
{
    print STDERR "Switches --training and --test should be used simultaneously.\n";
    askHelp();
    exit;
}

# pull in the regex filename and the xml source file, if any
$regexFile = shift;
$xmlSource = shift;

# try to open the source and destination file(s)
if ( defined $opt_training )
{ 
    open TRAIN_SRC, $opt_training || die "Couldnt open source file $opt_training\n";

    my $temp = $opt_training . ".arff";
    open TRAIN_DST, ">$temp" || die "Couldnt open destination file $temp\n";
    print TRAIN_DST "\@relation \'$relationName\'\n";
    
    open TEST_SRC, $opt_test || die "Couldnt open source file $opt_test\n";

    $temp = $opt_test . ".arff";
    open TEST_DST, ">$temp" || die "Couldnt open destination file $temp\n";
    print TEST_DST "\@relation \'$relationName\'\n";

    if ( defined $xmlSource ) { print "Ignoring redundant source file $xmlSource.\n"; }
}
else
{
    if ( !defined $xmlSource ) 
    {
	print "No source provided.\n";
	askHelp();
	exit;
    }
    open SRC, $xmlSource || die "Couldnt open source file $xmlSource\n";
    

    my $temp = $xmlSource . ".arff";
    open DST, ">$temp" || die "Couldnt open destination file $temp\n";
    print DST "\@relation \'$relationName\'\n";
}

# having done that, open the regex file
open REGEX, $regexFile || die "Couldnt open regex file $regexFile\n";

$i = 0;
while ( <REGEX> ) 
{
    chomp;
    ($regexArray[$i], $attribName) = (/\@name/)?(split/\s*\@name\s*=?\s*/):($_, "attribute-$i");
    
    # remove leading/trailing spaces
    $regexArray[$i] =~ s/^\s*(.*\S)?\s*$/$1/;

    # check for initial and ending / mark. otherwise invalid regex
    if ( $regexArray[$i] =~ s/^\/(.*)\/$/$1/ )
    {
	$i++;
    }
    else { next; } # current line is not valid regex!

    # escape the ' character in the attribName
    $attribName =~ s/'/\\'/g;

    if ( defined $opt_training )
    {
	print TRAIN_DST "\@attribute \'$attribName\' \{0,1\}\n";
	print TEST_DST "\@attribute \'$attribName\' \{0,1\}\n";
    }
    else { print DST "\@attribute \'$attribName\' \{0,1\}\n"; }
}
close REGEX;

# now go thru the lex file(s) to create the classString
if ( defined $opt_training )
{
    while ( <TRAIN_SRC> )
    {
	if ( /^<answer instance=\"[^\"]*\" senseid=\"([^\"]*)\"/ )
	{
	    $answer = $1;
	    $outputClass{$answer} = 1;
	}
    }
    close TRAIN_SRC;
    open TRAIN_SRC, $opt_training || die "Couldnt open file $opt_training\n";

    while ( <TEST_SRC> )
    {
	if ( /^<answer instance=\"[^\"]*\" senseid=\"([^\"]*)\"/ )
	{
	    $answer = $1;
	    $outputClass{$answer} = 1;
	}
    }
    close TEST_SRC;
    open TEST_SRC, $opt_test || die "Couldnt open file $opt_test\n";
}
else 
{
    while ( <SRC> )
    {
	if ( /^<answer instance=\"[^\"]*\" senseid=\"([^\"]*)\"/ )
	{
	    $answer = $1;
	    $outputClass{$answer} = 1;
	}
    }
    close SRC;
    open SRC, $xmlSource || die "Couldnt open file $xmlSource\n";
}

if ( ! defined %outputClass ) 
{
    print "Couldnt find any output classes... aborting!\n";
    exit;
}
else
{
    # ---------
    # ADP.3.2 
    # ---------
    # changing the name of the last attribute to 'senseclass' instead of 
    # 'class' which frequently occurs as an attribute and causes problem 
    # for a program that checks for duplicate attributes 
    # version 0.1 code
    #$classString = "\@attribute \'class\' \{";

    $classString = "\@attribute \'senseclass\' \{";
    my $firstTime = 1;
    foreach ( keys %outputClass ) 
    { 
	if ( !($firstTime) ) { $classString .= ", "; }
	else { $firstTime = 0; }
	
	$classString .= "$_";
    }
    $classString .= "\}\n";
}

# having created the class string, print it out where necessary
if ( defined $opt_training )
{
    print TRAIN_DST $classString;
    print TRAIN_DST "\@data\n";
    print TEST_DST $classString;
    print TEST_DST "\@data\n";
}
else { print DST $classString; print DST "\@data\n"; }

# now to create the actual arff(s)!!
	
$continue = 1; # if --training, then continue will go 1->2->0. else it will go 1->0
while ( $continue )
{
    $insideContext = 0;
    $line = "";
    $done = 0;

    while ( !$done )
    {
	my $nextLine = ($continue == 2) ? <TEST_SRC> : ( (defined $opt_training) ? <TRAIN_SRC> : <SRC> );
       	if ( !defined $nextLine ) { $done = 1; next; } # end of file... so quit!

	if ( $insideContext ) 
	{
	    if ( $nextLine =~ /^<\/context>/ )
	    {
		$outputData = "";

		# for each answer senseid, do the following
		if ( $#answers < 0 ) { push @answers, '?'; }
		foreach $answer (@answers)
		{
		    $firstOne = 1;
		    $outputData = "\{";
		    # now to get the sparse output done!
		    for ( $i = 0; $i <= $#regexArray; $i++ )
		    {
#  			# from the regex, find out the tags to keep in the context
#  			@keepTags = ();
#  			while ( $regexArray[$i] =~ /<[^>]*>/g ) 
#  			{ 
#  			    push @keepTags, $&; 
#  			}
			
#  			# now remove all tags from $line that we dont need
#  			$newLine = "";
#  			$temp = $line;
			
#  			while ( $temp =~ /<[^>]*>/ )
#  			{
#  			    $newLine .= $`;
#  			    $temp = $';
#  			    foreach $tag (@keepTags) 
#  			    {
#  				if ( $& =~ /^$tag$/ ) 
#  				{ 
#  				    $newLine .= $&; 
#  				    last;
#  				}
#  			    }
#  			}
			
#  			$newLine .= $temp;
			
			if ( $line =~ $regexArray[$i] )
			{
			    if ( !($firstOne) ) { $outputData .= ", "; }
			    else { $firstOne = 0; }
			    
			    $outputData .= "$i 1";
			}
		    }
		    if ( !( $firstOne ) ) { $outputData .= ", "; }
		    # version 0.1 code
		    #$outputData .= "$i $answer\} % $lexelt $word.$instanceId\n";
		    # this code assumed that lexelt has . in it and 
		    # we can separate word which is part of instance id 
		    # instead we take instance id and lexelt as they appear 
		    # --------
		    # ADP.3.1
		    # --------
		    $outputData .= "$i $answer\} % $lexelt $instanceId\n";	

		    if ( $continue == 2 ) 
		    {
			print TEST_DST $outputData;
			last;
		    }
		    elsif ( defined $opt_training )
		    {
			print TRAIN_DST $outputData;
		    }
		    else 
		    {
			print DST $outputData;
		    }
		}
		
		$insideContext = 0;
		@answers = ();
		$line = "";
		next;
	    }
	    
	    $line .= $nextLine;
	}
	else
	{
	    if ( $nextLine =~ /^<answer instance=\"[^\"]*\" senseid=\"([^\"]*)\"/ )
	    {
		$answer = $1;
		push @answers, $answer;
		next;
	    }
	    
	    if ( $nextLine =~ /^<context>/ ) { $insideContext = 1; }

#	    commented code belongs to version 0.1 	
#	    this code assumed that instance id has . in it 
#	    and may not work for data other than Senseval
#	    instead now it will accept the instance id as it appears 
#	    if ( $nextLine =~ /^<instance id=\"[^.]*\.(\d+)\"/) 

	    # --------
            # ADP.3.1
            # --------
	    if($nextLine =~ /^<instance id=\"([^"]*)\"/)
	    {
		$instanceId = $1;
	    }
	    
	    if ( $nextLine =~ /^<lexelt item=\"([^\"]*)\"/ )
	    {
		$lexelt = $1;

		# lexelt could now be word.suffix. get rid of suffix

		# verion 0.1 code
		#my $suffix;
		# this code assumed that the lexelt element had . in it 
		# which may not work for data other than senseval
		# --------
                # ADP.3.1
                # --------
		#($word, $suffix) = split(/\./, $lexelt);		
	    }
	}
    }

    if ( $continue == 2 ) { $continue = 0; }
    elsif ( defined $opt_training ) { $continue = 2; }
    else { $continue = 0; }
}

close();

# thats it!!

# function to output a minimal usage note when the user has not provided any
# commandline options
sub minimalUsageNotes
{
    print "Usage: xml2arff.pl [OPTIONS] REGEX_FILE XML_SOURCE\n";
    askHelp();
}

# function to output help messages for this program
sub showHelp
{
    print "Usage: xml2arff.pl [OPTIONS] REGEX_FILE XML_SOURCE\n\n";

    print "Converts xml-format Senseval file XML_SOURCE to arff file XML_SOURCE.arff\n";
    print "based on given regular expressions in REGEX_FILE.\n\n";

    print "OPTIONS:\n\n";

    print "  --version          Prints the version number.\n\n";

    print "  --help             Prints this help message.\n\n";

    print "  --relation NAME    Name of the relation in the arff file. \"RELATION\"\n";
    print "                     by default.\n\n";

    print "  --training FILE    Name of xml file, instances in which are to be used\n";
    print "                     as training data. Resulting arff file output to FILE.arff\n";
    print "                     Must be provided if --test is used.\n\n";

    print "  --test FILE        Name of xml file, instances in which are to be used\n";
    print "                     as test data. Resulting arff file output to FILE.arff\n";
    print "                     Must be provided if --training is used.\n\n";

    print "Note: If --training and --test are provided, XML_SOURCE will be ignored.\n";
    print "      \@attribute 'class' will consist of all unique class labels across\n";
    print "      across both --training file and --test file.\n\n";

}

# function to output the version number
sub showVersion
{
    print "xml2arff.pl  -  Version 0.3\n";
    print "A component of SenseTools 0.3\n";
    print "Copyright (C) 2001-2003, Ted Pedersen & Satanjeev Banerjee\n";
    print "Date of Last Update: May 01, 2003 by TDP\n";

}

# function to output "ask for help" message when the user's goofed up!
sub askHelp 
{
    print "Type xml2arff.pl --help for help.\n";
}
